/*
# mk_unset.c
# Copyright (C) 2023 Joseph Rosevear, San Diego CA, USA.

# This file is part of a distribution SAM.

# SAM is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your
# option) any later version.

# SAM is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



# Joseph Rosevear I wrote this source file.*/



#include <stdio.h>

int main() {

   int c;
   char zone, instring;

   zone=0;
   instring=0;
   while ((c = getchar()) != EOF ) {

      /* Zone 1 is the first part that you get from set.  It says "blah = something",
      but zone 1 is is just "blah".  So zone 1 ends when "=" comes.                    */
      switch(zone) {
      case 1:
         
         switch(c) {
         case '=':
            zone = (++zone % 3);
            /* This allows "set" to be run without annoying error messages which result
            from setting read only variables.  I decided to do this instead of using the
            output of the "readonly" command to filter out the setting of the readonly's.*/
            printf(" 2>/dev/null");
            break;
         /* Don't allow a new line before the "=" */
         case '\n':
            fprintf(stderr, "Premature end of line found in input file\n");
            return(1);
            break;
         default:
            putchar(c);
         }

      /* Zone 2 is the "something" from above.  It doesn't get used by this program except
      to signal when zone 0 begins.                                                       */
      case 2:

         switch(instring) {
         /* Zone 2 ends when new line happens, but only when not in a string. */ 
         case 0:
            if (c == '\n') {
               zone = (++zone % 3);
               putchar(c);
            }
            if (c == '\'') instring = 1;
            break;
         case 1:
            if (c == '\'') instring = 0;
            break;
         }   

         break;
      /* Zone 0 is the waiting for a new "blah = something".  It was the initial zone,
      and it comes after each zone 2.                                                 */
      case 0:
         if (c != '\n') {
            zone = (++zone % 3);
            printf("unset ");
            putchar(c);
         }
         break;
      }
   }

   /* When we get here end-of-file has been encountered. */
   switch(zone) {
   /* Ideally we should be in zone 0.  If we're in zone 1, it means "blah = something" was
   begun but EOF came before the "=".  That's an error.                                  */
   case 1:
      fprintf(stderr, "Premature end of file found in input file\n");
      return(1);
      break;
   /*If we're in zone 2, it means the "=" occurred, and we hope "something" came after.
   We don't know.  The newline hasn't come yet or we would be in zone 0, so we'll supply
   it.                                                                                  */
   case 2:
      /* missing final \n; will supply it */
      putchar('\n');
      break;
   }

   /* We should not be in a string when EOF comes.                                      */
   if (instring) {
      fprintf(stderr, "No matching \' found in input file\n");
      return(1);
   }

   return(0);
}   
